@@ -24,7 +24,7 @@ from urllib.request import Request, urlopen, urljoin, quote
 from urllib.error import URLError
 
 from jupyter_core.paths import jupyter_config_path
-from jupyterlab_server.process import which, Process, WatchHelper
+from jupyterlab_server.process import which, Process, WatchHelper, list2cmdline
 from notebook.nbextensions import GREEN_ENABLED, GREEN_OK, RED_DISABLED, RED_X
 
 from .semver import Range, gte, lt, lte, gt, make_semver
@@ -35,13 +35,77 @@ from .jlpmapp import YARN_PATH, HERE
 WEBPACK_EXPECT = re.compile(r'.*/index.out.js')
 
 # The dev mode directory.
-DEV_DIR = osp.realpath(os.path.join(HERE, '..', 'dev_mode'))
+DEV_DIR = osp.abspath(os.path.join(HERE, '..', 'dev_mode'))
+
+
+class ProgressProcess(Process):
+
+    def __init__(self, cmd, logger=None, cwd=None, kill_event=None,
+                 env=None):
+        """Start a subprocess that can be run asynchronously.
+
+        Parameters
+        ----------
+        cmd: list
+            The command to run.
+        logger: :class:`~logger.Logger`, optional
+            The logger instance.
+        cwd: string, optional
+            The cwd of the process.
+        kill_event: :class:`~threading.Event`, optional
+            An event used to kill the process operation.
+        env: dict, optional
+            The environment for the process.
+        """
+        if not isinstance(cmd, (list, tuple)):
+            raise ValueError('Command must be given as a list')
+
+        if kill_event and kill_event.is_set():
+            raise ValueError('Process aborted')
+
+        self.logger = logger = logger or logging.getLogger('jupyterlab')
+        self._last_line = ''
+        self.cmd = cmd
+        self.logger.debug('> ' + list2cmdline(cmd))
+
+        self.proc = self._create_process(
+            cwd=cwd,
+            env=env,
+            stderr=subprocess.STDOUT,
+            stdout=subprocess.PIPE,
+            universal_newlines=True,
+        )
+        self._kill_event = kill_event or Event()
+
+        Process._procs.add(self)
+
+    def wait(self):
+        cache = []
+        proc = self.proc
+        kill_event = self._kill_event
+        import itertools
+        spinner = itertools.cycle(['-', '/', '|', '\\'])
+        while proc.poll() is None:
+            sys.stdout.write(next(spinner))   # write the next character
+            sys.stdout.flush()                # flush stdout buffer (actual character display)
+            sys.stdout.write('\b')
+            if kill_event.is_set():
+                self.terminate()
+                raise ValueError('Process was aborted')
+            try:
+                out, _ = proc.communicate(timeout=.1)
+                cache.append(out)
+            except subprocess.TimeoutExpired:
+                continue
+        self.logger.debug('\n'.join(cache))
+        sys.stdout.flush()
+        return self.terminate()
 
 
 def pjoin(*args):
     """Join paths to create a real path.
     """
-    return osp.realpath(osp.join(*args))
+    return osp.abspath(osp.join(*args))
 
 
 def get_user_settings_dir():
@@ -51,7 +115,7 @@ def get_user_settings_dir():
     settings_dir = settings_dir or pjoin(
         jupyter_config_path()[0], 'lab', 'user-settings'
     )
-    return osp.realpath(settings_dir)
+    return osp.abspath(settings_dir)
 
 
 def get_workspaces_dir():
@@ -61,7 +125,7 @@ def get_workspaces_dir():
     workspaces_dir = workspaces_dir or pjoin(
         jupyter_config_path()[0], 'lab', 'workspaces'
     )
-    return osp.realpath(workspaces_dir)
+    return osp.abspath(workspaces_dir)
 
 
 def get_app_dir():
@@ -69,7 +133,7 @@ def get_app_dir():
     """
     # Default to the override environment variable.
     if os.environ.get('JUPYTERLAB_DIR'):
-        return osp.realpath(os.environ['JUPYTERLAB_DIR'])
+        return osp.abspath(os.environ['JUPYTERLAB_DIR'])
 
     # Use the default locations for data_files.
     app_dir = pjoin(sys.prefix, 'share', 'jupyter', 'lab')
@@ -87,29 +151,56 @@ def get_app_dir():
           osp.exists(app_dir) and
           osp.exists('/usr/local/share/jupyter/lab')):
         app_dir = '/usr/local/share/jupyter/lab'
+    return osp.abspath(app_dir)
 
-    return osp.realpath(app_dir)
+
+def dedupe_yarn(path, logger=None):
+    """ `yarn-deduplicate` with the `fewer` strategy to minimize total
+        packages installed in a given staging directory
+
+        This means a extension (or dependency) _could_ cause a downgrade of an
+        version expected at publication time, but core should aggressively set
+        pins above, for example, known-bad versions
+    """
+    had_dupes = ProgressProcess(
+        ['node', YARN_PATH, 'yarn-deduplicate', '-s', 'fewer'],
+        cwd=path, logger=logger
+    ).wait() != 0
+
+    if had_dupes:
+        yarn_proc = ProgressProcess(['node', YARN_PATH], cwd=path, logger=logger)
+        yarn_proc.wait()
+
+
+def ensure_node_modules(cwd, logger=None):
+    """Ensure that node_modules is up to date.
+
+    Returns true if the node_modules was updated.
+    """
+    logger = _ensure_logger(logger)
+    yarn_proc = ProgressProcess(['node', YARN_PATH, 'check', '--verify-tree'], cwd=cwd, logger=logger)
+    ret = yarn_proc.wait()
+
+    # Update node_modules if needed.
+    if ret != 0:
+        yarn_proc = ProgressProcess(['node', YARN_PATH], cwd=cwd, logger=logger)
+        yarn_proc.wait()
+        parent = pjoin(HERE, '..')
+        dedupe_yarn(parent, logger)
+
+    return ret != 0
 
 
 def ensure_dev(logger=None):
     """Ensure that the dev assets are available.
     """
     parent = pjoin(HERE, '..')
+    logger = _ensure_logger(logger)
+    target = pjoin(parent, 'dev_mode', 'static')
 
-    if not osp.exists(pjoin(parent, 'node_modules')):
-        yarn_proc = Process(['node', YARN_PATH], cwd=parent, logger=logger)
-        yarn_proc.wait()
-
-    theme_packages = ['theme-light-extension', 'theme-dark-extension']
-    for theme in theme_packages:
-        base_path = pjoin(parent, 'packages', theme)
-        if not osp.exists(pjoin(base_path, 'static')):
-            yarn_proc = Process(['node', YARN_PATH, 'build:webpack'],
-                                cwd=base_path, logger=logger)
-            yarn_proc.wait()
-
-    if not osp.exists(pjoin(parent, 'dev_mode', 'static')):
-        yarn_proc = Process(['node', YARN_PATH, 'build'], cwd=parent,
+    # Determine whether to build.
+    if ensure_node_modules(parent, logger) or not osp.exists(target):
+        yarn_proc = ProgressProcess(['node', YARN_PATH, 'build'], cwd=parent,
                             logger=logger)
         yarn_proc.wait()
 
@@ -118,21 +209,30 @@ def ensure_core(logger=None):
     """Ensure that the core assets are available.
     """
     staging = pjoin(HERE, 'staging')
+    logger = _ensure_logger(logger)
 
-    # Bail if the static directory already exists.
-    if osp.exists(pjoin(HERE, 'static')):
-        return
-
-    if not osp.exists(pjoin(staging, 'node_modules')):
-        yarn_proc = Process(['node', YARN_PATH], cwd=staging, logger=logger)
-        yarn_proc.wait()
-
-    if not osp.exists(pjoin(HERE, 'static')):
-        yarn_proc = Process(['node', YARN_PATH, 'build'], cwd=staging,
+    # Determine whether to build.
+    target = pjoin(HERE, 'static', 'index.html')
+    if not osp.exists(target):
+        ensure_node_modules(staging, logger)
+        yarn_proc = ProgressProcess(['node', YARN_PATH, 'build'], cwd=staging,
                             logger=logger)
         yarn_proc.wait()
 
 
+def ensure_app(app_dir):
+    """Ensure that an application directory is available.
+
+    If it does not exist, return a list of messages to prompt the user.
+    """
+    if osp.exists(pjoin(app_dir, 'static', 'index.html')):
+        return
+
+    msgs = ['JupyterLab application assets not found in "%s"' % app_dir,
+            'Please run `jupyter lab build` or use a different app directory']
+    return msgs
+
+
 def watch_packages(logger=None):
     """Run watch mode for the source packages.
 
@@ -146,13 +246,10 @@ def watch_packages(logger=None):
     A list of `WatchHelper` objects.
     """
     parent = pjoin(HERE, '..')
-
-    if not osp.exists(pjoin(parent, 'node_modules')):
-        yarn_proc = Process(['node', YARN_PATH], cwd=parent, logger=logger)
-        yarn_proc.wait()
-
     logger = _ensure_logger(logger)
-    ts_dir = osp.realpath(osp.join(HERE, '..', 'packages', 'metapackage'))
+    ensure_node_modules(parent, logger)
+
+    ts_dir = osp.abspath(osp.join(HERE, '..', 'packages', 'metapackage'))
 
     # Run typescript watch and wait for the string indicating it is done.
     ts_regex = r'.* Found 0 errors\. Watching for file changes\.'
@@ -219,7 +316,7 @@ def install_extension(extension, app_dir=None, logger=None):
     return handler.install_extension(extension)
 
 
-def uninstall_extension(name, app_dir=None, logger=None):
+def uninstall_extension(name=None, app_dir=None, logger=None, all_=False):
     """Uninstall an extension by name or path.
 
     Returns `True` if a rebuild is recommended, `False` otherwise.
@@ -227,6 +324,8 @@ def uninstall_extension(name, app_dir=None, logger=None):
     logger = _ensure_logger(logger)
     _node_check(logger)
     handler = _AppHandler(app_dir, logger)
+    if all_ is True:
+        return handler.uninstall_all_extensions()
     return handler.uninstall_extension(name)
 
 
@@ -262,7 +361,7 @@ def clean(app_dir=None, logger=None):
     logger.info('Success!')
 
 
-def build(app_dir=None, name=None, version=None, public_url=None,
+def build(app_dir=None, name=None, version=None, static_url=None,
           logger=None, command='build:prod', kill_event=None,
           clean_staging=False):
     """Build the JupyterLab application.
@@ -270,7 +369,7 @@ def build(app_dir=None, name=None, version=None, public_url=None,
     logger = _ensure_logger(logger)
     _node_check(logger)
     handler = _AppHandler(app_dir, logger, kill_event=kill_event)
-    return handler.build(name=name, version=version, public_url=public_url,
+    return handler.build(name=name, version=version, static_url=static_url,
                          command=command, clean_staging=clean_staging)
 
 
@@ -434,25 +533,37 @@ class _AppHandler(object):
 
         return True
 
-    def build(self, name=None, version=None, public_url=None,
+    def build(self, name=None, version=None, static_url=None,
               command='build:prod', clean_staging=False):
         """Build the application.
         """
+        self.logger.info('Building jupyterlab assets')
+
         # Set up the build directory.
         app_dir = self.app_dir
 
         self._populate_staging(
-            name=name, version=version, public_url=public_url,
+            name=name, version=version, static_url=static_url,
             clean=clean_staging
         )
 
         staging = pjoin(app_dir, 'staging')
 
         # Make sure packages are installed.
-        self._run(['node', YARN_PATH, 'install'], cwd=staging)
+        ret = self._run(['node', YARN_PATH, 'install', '--non-interactive'], cwd=staging)
+        if ret != 0:
+            msg = 'npm dependencies failed to install'
+            self.logger.debug(msg)
+            raise RuntimeError(msg)
+
+        dedupe_yarn(staging, self.logger)
 
         # Build the app.
-        self._run(['node', YARN_PATH, 'run', command], cwd=staging)
+        ret = self._run(['node', YARN_PATH, 'run', command], cwd=staging)
+        if ret != 0:
+            msg = 'JupyterLab failed to build'
+            self.logger.debug(msg)
+            raise RuntimeError(msg)
 
     def watch(self):
         """Start the application watcher and then run the watch in
@@ -464,6 +575,7 @@ class _AppHandler(object):
 
         # Make sure packages are installed.
         self._run(['node', YARN_PATH, 'install'], cwd=staging)
+        dedupe_yarn(staging, self.logger)
 
         proc = WatchHelper(['node', YARN_PATH, 'run', 'watch'],
                            cwd=pjoin(self.app_dir, 'staging'),
@@ -477,43 +589,43 @@ class _AppHandler(object):
         logger = self.logger
         info = self.info
 
-        logger.info('JupyterLab v%s' % info['version'])
+        print('JupyterLab v%s' % info['version'])
 
         if info['extensions']:
             info['compat_errors'] = self._get_extension_compat()
-            logger.info('Known labextensions:')
+            print('Known labextensions:')
             self._list_extensions(info, 'app')
             self._list_extensions(info, 'sys')
         else:
-            logger.info('No installed extensions')
+            print('No installed extensions')
 
         local = info['local_extensions']
         if local:
-            logger.info('\n   local extensions:')
+            print('\n   local extensions:')
             for name in sorted(local):
-                logger.info('        %s: %s' % (name, local[name]))
+                print('        %s: %s' % (name, local[name]))
 
         linked_packages = info['linked_packages']
         if linked_packages:
-            logger.info('\n   linked packages:')
+            print('\n   linked packages:')
             for key in sorted(linked_packages):
                 source = linked_packages[key]['source']
-                logger.info('        %s: %s' % (key, source))
+                print('        %s: %s' % (key, source))
 
         uninstalled_core = info['uninstalled_core']
         if uninstalled_core:
-            logger.info('\nUninstalled core extensions:')
-            [logger.info('    %s' % item) for item in sorted(uninstalled_core)]
+            print('\nUninstalled core extensions:')
+            [print('    %s' % item) for item in sorted(uninstalled_core)]
 
         disabled_core = info['disabled_core']
         if disabled_core:
-            logger.info('\nDisabled core extensions:')
-            [logger.info('    %s' % item) for item in sorted(disabled_core)]
+            print('\nDisabled core extensions:')
+            [print('    %s' % item) for item in sorted(disabled_core)]
 
         messages = self.build_check(fast=True)
         if messages:
-            logger.info('\nBuild recommended, please run `jupyter lab build`:')
-            [logger.info('    %s' % item) for item in messages]
+            print('\nBuild recommended, please run `jupyter lab build`:')
+            [print('    %s' % item) for item in messages]
 
     def build_check(self, fast=False):
         """Determine whether JupyterLab should be built.
@@ -623,6 +735,17 @@ class _AppHandler(object):
         self.logger.warn('No labextension named "%s" installed' % name)
         return False
 
+    def uninstall_all_extensions(self):
+        """Uninstalls all extensions
+
+        Returns `True` if a rebuild is recommended, `False` otherwise
+        """
+        should_rebuild = False
+        for (extname, _) in self.info['extensions'].items():
+            uninstalled = self.uninstall_extension(extname)
+            should_rebuild = should_rebuild or uninstalled
+        return should_rebuild
+
     def update_all_extensions(self):
         """Update all non-local extensions.
 
@@ -657,6 +780,7 @@ class _AppHandler(object):
         except URLError:
             return False
         if latest is None:
+            self.logger.warn('No compatible version found for %s!' % (name,))
             return False
         if latest == self.info['extensions'][name]['version']:
             self.logger.info('Extension %r already up to date' % name)
@@ -825,7 +949,7 @@ class _AppHandler(object):
         info['static_data'] = _get_static_data(self.app_dir)
         app_data = info['static_data'] or core_data
         info['version'] = app_data['jupyterlab']['version']
-        info['publicUrl'] = app_data['jupyterlab'].get('publicUrl', '')
+        info['staticUrl'] = app_data['jupyterlab'].get('staticUrl', '')
 
         info['sys_dir'] = self.sys_dir
         info['app_dir'] = self.app_dir
@@ -840,7 +964,7 @@ class _AppHandler(object):
         info['disabled_core'] = disabled_core
         return info
 
-    def _populate_staging(self, name=None, version=None, public_url=None,
+    def _populate_staging(self, name=None, version=None, static_url=None,
                           clean=False):
         """Set up the assets in the staging directory.
         """
@@ -870,12 +994,6 @@ class _AppHandler(object):
             target = pjoin(staging, fname)
             shutil.copy(pjoin(HERE, 'staging', fname), target)
 
-        # Remove an existing yarn.lock file
-        # Because otherwise we can end up with unwanted duplicates
-        # cf https://github.com/yarnpkg/yarn/issues/3967
-        if osp.exists(pjoin(staging, 'yarn.lock')):
-            os.remove(pjoin(staging, 'yarn.lock'))
-
         # Ensure a clean templates directory
         templates = pjoin(staging, 'templates')
         if osp.exists(templates):
@@ -935,13 +1053,18 @@ class _AppHandler(object):
         if name:
             data['jupyterlab']['name'] = name
 
-        if public_url:
-            data['jupyterlab']['publicUrl'] = public_url
+        if static_url:
+            data['jupyterlab']['staticUrl'] = static_url
 
         pkg_path = pjoin(staging, 'package.json')
         with open(pkg_path, 'w') as fid:
             json.dump(data, fid, indent=4)
 
+        # copy known-good yarn.lock if missing
+        lock_path = pjoin(staging, 'yarn.lock')
+        if not osp.exists(lock_path):
+            shutil.copy(pjoin(HERE, 'staging', 'yarn.lock'), lock_path)
+
     def _get_package_template(self, silent=False):
         """Get the template the for staging package.json file.
         """
@@ -1074,7 +1197,7 @@ class _AppHandler(object):
             deps = data.get('dependencies', dict())
             name = data['name']
             jlab = data.get('jupyterlab', dict())
-            path = osp.realpath(target)
+            path = osp.abspath(target)
             # homepage, repository  are optional
             if 'homepage' in data:
                 url = data['homepage']
@@ -1119,7 +1242,7 @@ class _AppHandler(object):
             return info
 
         for path in glob.glob(pjoin(dname, '*.tgz')):
-            path = osp.realpath(path)
+            path = osp.abspath(path)
             data = read_package(path)
             name = data['name']
             if name not in info:
@@ -1277,8 +1400,8 @@ class _AppHandler(object):
                     raise ValueError(msg)
 
                 if version and name:
-                    self.logger.warning('Incompatible extension:\n%s', msg)
-                    self.logger.warning('Found compatible version: %s', version)
+                    self.logger.debug('Incompatible extension:\n%s', name)
+                    self.logger.debug('Found compatible version: %s', version)
                     with TemporaryDirectory() as tempdir2:
                         return self._install_extension(
                             '%s@%s' % (name, version), tempdir2)
@@ -1302,7 +1425,7 @@ class _AppHandler(object):
         info['path'] = target
         return info
 
-    def _extract_package(self, source, tempdir, quiet=False):
+    def _extract_package(self, source, tempdir):
         """Call `npm pack` for an extension.
 
         The pack command will download the package tar if `source` is
@@ -1311,11 +1434,11 @@ class _AppHandler(object):
         """
         is_dir = osp.exists(source) and osp.isdir(source)
         if is_dir and not osp.exists(pjoin(source, 'node_modules')):
-            self._run(['node', YARN_PATH, 'install'], cwd=source, quiet=quiet)
+            self._run(['node', YARN_PATH, 'install'], cwd=source)
 
         info = dict(source=source, is_dir=is_dir)
 
-        ret = self._run([which('npm'), 'pack', source], cwd=tempdir, quiet=quiet)
+        ret = self._run([which('npm'), 'pack', source], cwd=tempdir)
         if ret != 0:
             msg = '"%s" is not a valid npm package'
             raise ValueError(msg % source)
@@ -1357,10 +1480,17 @@ class _AppHandler(object):
             errors = _validate_compatibility(name, deps, core_data)
             if not errors:
                 # Found a compatible version
+                # skip deprecated versions
+                if 'deprecated' in data:
+                    self.logger.debug(
+                        'Disregarding compatible version of package as it is deprecated: %s@%s'
+                        % (name, version)
+                    )
+                    continue
                 # Verify that the version is a valid extension.
                 with TemporaryDirectory() as tempdir:
                     info = self._extract_package(
-                        '%s@%s' % (name, version), tempdir, quiet=True)
+                        '%s@%s' % (name, version), tempdir)
                 if _validate_extension(info['data']):
                     # Invalid, do not consider other versions
                     return
@@ -1390,6 +1520,11 @@ class _AppHandler(object):
             for version, data in sorted(versions.items(),
                                         key=sort_key,
                                         reverse=True):
+
+                # skip deprecated versions
+                if 'deprecated' in data:
+                    continue
+
                 deps = data.get('dependencies', {})
                 errors = _validate_compatibility(name, deps, core_data)
                 if not errors:
@@ -1402,7 +1537,7 @@ class _AppHandler(object):
         if not keys:
             return versions
         with TemporaryDirectory() as tempdir:
-            ret = self._run([which('npm'), 'pack'] + keys, cwd=tempdir, quiet=True)
+            ret = self._run([which('npm'), 'pack'] + keys, cwd=tempdir)
             if ret != 0:
                 msg = '"%s" is not a valid npm package'
                 raise ValueError(msg % keys)
@@ -1413,7 +1548,7 @@ class _AppHandler(object):
                 # Verify that the version is a valid extension.
                 if not _validate_extension(data):
                     # Valid
-                    versions[key] = data['version']
+                    versions[data['name']] = data['version']
         return versions
 
     def _format_no_compatible_package_version(self, name):
@@ -1471,7 +1606,7 @@ class _AppHandler(object):
 
         kwargs['logger'] = self.logger
         kwargs['kill_event'] = self.kill_event
-        proc = Process(cmd, **kwargs)
+        proc = ProgressProcess(cmd, **kwargs)
         return proc.wait()
 
 
@@ -1481,7 +1616,7 @@ def _node_check(logger):
     node = which('node')
     try:
         output = subprocess.check_output([node, 'node-version-check.js'], cwd=HERE)
-        logger.info(output.decode('utf-8'))
+        logger.debug(output.decode('utf-8'))
     except Exception:
         data = _get_core_data()
         ver = data['engines']['node']
@@ -1523,7 +1658,7 @@ def _validate_extension(data):
         return ['The `jupyterlab` key must be a JSON object']
     extension = jlab.get('extension', False)
     mime_extension = jlab.get('mimeExtension', False)
-    themeDir = jlab.get('themeDir', '')
+    themePath = jlab.get('themePath', '')
     schemaDir = jlab.get('schemaDir', '')
 
     messages = []
@@ -1555,8 +1690,8 @@ def _validate_extension(data):
     if mime_extension and mime_extension not in files:
         messages.append('Missing mimeExtension module "%s"' % mime_extension)
 
-    if themeDir and not any(f.startswith(themeDir) for f in files):
-        messages.append('themeDir is empty: "%s"' % themeDir)
+    if themePath and not any(f.startswith(themePath) for f in files):
+        messages.append('themePath is empty: "%s"' % themePath)
 
     if schemaDir and not any(f.startswith(schemaDir) for f in files):
         messages.append('schemaDir is empty: "%s"' % schemaDir)
